%%writefile test.cu
#include <stdlib.h>
#include <stdio.h>
#include "time.h"

#define NMAX 100
#define DATAMAX 1000
#define DATAMIN -1000

__host__ __device__ void print_matrix(int *m,int row, int col)
{
  for (int i = 0; i < row; i++)
  {
    for (int j = 0; j < col; j++)
    {
      printf("%d ", m[i * col + j]);
    }
    printf("\n");
  }
}


__device__ int supression_op(int *kernel, int *target, int rowKernel,int colKernel,int rowTarget,int colTarget, int iterationRow, int iterationCol)
{
  int intermediate_sum = 0;
  for (int i = 0; i < rowKernel; i++)
  {
    for (int j = 0; j < colKernel; j++)
    {
      intermediate_sum += kernel[i * colKernel + j] * target[(iterationRow + i) * colTarget + iterationCol + j];
    }
  }

  return intermediate_sum;
}

__global__ void convolution(int *output,int *kernel, int *target, int rowKernel, int colKernel, int rowTarget, int colTarget, int index)
{
  int rowC = rowTarget-rowKernel + 1;
  int colC = colTarget-colKernel + 1;

   // BLOCK
  int block = blockIdx.x;
  int numberOfBlock = gridDim.x;
  // printf("NUMBER OF BLOCK BLOCK %d \n",numberOfBlock);
  // int k = block;
  // printf("BLOCK %d \n",blockIdx.x);
  // THREAD

  int tx= threadIdx.x;
  int ty= threadIdx.y;

  // int T = blockIdx.x * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x; // ALL THREAD INDEX
  
  for (int k = block; k < index; k+= numberOfBlock){
    for (int i = tx; i < rowC; i+= blockDim.x){
      for (int j = ty; j < colC; j += blockDim.y){
        // printf("BLOCK %d ITERATION k : %d THREAD %d FOR i : %d FOR j : %d \n",block,T,k,i,j);
        (&output[k*rowC*colC])[i*colC + j] = supression_op(kernel, (&target[k*rowTarget*colTarget]), rowKernel, colKernel, rowTarget, colTarget, i, j);
      }
    }
  }
}

__device__ int getDataRange(int *source, int rowA, int colA){
  int max = DATAMIN;
  int min = DATAMAX;
  for (int i = 0; i < rowA; i++)
  {
    for (int j = 0; j < colA; j++)
    {
      int el = source[i * colA + j];
      if (el > max)
        max = el;
      if (el < min)
        min = el;
    }
  }

  return max - min;
}
__global__ void getDataRangeArray(int *output, int *source, int index,int rowSource, int colSource){
  int tx = threadIdx.x;
  int threadXinBlock = blockDim.x;
  // printf("output : %d, i : %d \n",getDataRange(&source[-1*rowSource*colSource],rowSource,colSource),0);
  // printf("output : %d, i : %d \n",getDataRange(&source[0*rowSource*colSource],rowSource,colSource),0);
  // printf("output : %d, i : %d \n",getDataRange(&source[1*rowSource*colSource],rowSource,colSource),1);
  // printf("output : %d, i : %d \n",getDataRange(&source[2*rowSource*colSource],rowSource,colSource),2);
  for( int i = tx ; i < index ; i+=threadXinBlock ){
    output[i] = getDataRange(&source[i*rowSource*colSource], rowSource, colSource);
    // printf("output : %d, i : %d \n",output[i],i);
  }
  
}
__device__ void mergeArray(int *output, int *temporary_array, int left,int mid, int right){
  int iterleft = left, iterright = mid;
  for(int i = left; i < right; i++){
    if(iterleft < mid && (iterright >= right || output[iterleft] < output[iterright])){
      temporary_array[i] = output[iterleft++];
    }
    else{
      temporary_array[i] = output[iterright++];
    }
  }
}

__global__ void mergeSort(int* output, int *temporary_array, int index, int width, int arrayPerThread){
  int x = blockIdx.x * blockDim.x + threadIdx.x;
  int left = width*x*arrayPerThread;
  int mid,right;

  for(int i = 0; i < arrayPerThread; i++){
    if(left >= index) break;

    mid = fminf(left+(width/2), index); 
    right = fminf(left+width, index);
    mergeArray(output, temporary_array, left, mid, right);
    left += width;
  }
}


void convolutionWithCUDA(int* output, int* kernel, int* target, int rowKernel, int colKernel, int rowTarget, int colTarget,int index, int block_num, int thread_num){
  // B = Kernel
  int* dev_a = nullptr;
  int* dev_b = nullptr;
  int* dev_c = nullptr;
  int* output_datarange = nullptr;
  int* temporary_array = nullptr;
  int* temp_swap;
  // int* datarange_sor = nullptr;
  // dim3* deviceBlock_num = nullptr;
  // dim3* deviceThread_num = nullptr;

  // Allocate GPU buffers for three vectors (two input, one output)
  // print_matrix(&target[0],3,3);
  // print_matrix(&target[1],3,3);
  // print_matrix(&target[2],3,3);
  // printf("\n %d \n", *(target[0][1]));
  int rowC = rowTarget-rowKernel + 1;
  int colC = colTarget-colKernel + 1;


  // cudaMalloc((void**)&deviceBlock_num, sizeof(dim3));
  // cudaMalloc((void**)&deviceThread_num, sizeof(dim3));
  
  cudaMalloc((void**)&output_datarange, index * sizeof(int));
  cudaMalloc((void**)&temporary_array, index * sizeof(int));
  cudaMalloc((void**)&dev_c,index * rowC * colC * sizeof(int));
  cudaMalloc((void**)&dev_a, rowKernel * colKernel * sizeof(int));
  cudaMalloc((void**)&dev_b,index * rowTarget * colTarget * sizeof(int));


  // cudaMemcpy(deviceBlock_num, block_num, sizeof(dim3), cudaMemcpyHostToDevice);
  // cudaMemcpy(deviceThread_num, thread_num, sizeof(dim3), cudaMemcpyHostToDevice);
  
  cudaMemcpy(dev_a, kernel, rowKernel * colKernel * sizeof(int), cudaMemcpyHostToDevice);
  cudaMemcpy(dev_b, target,index * rowTarget * colTarget * sizeof(int), cudaMemcpyHostToDevice);
  convolution<<<dim3(block_num,1,1),dim3(thread_num,thread_num,1)>>>(dev_c, dev_a, dev_b, rowKernel,colKernel,rowTarget,colTarget,index);
  cudaDeviceSynchronize();

  
  getDataRangeArray<<<1,thread_num>>>(output_datarange,dev_c,index,rowC,colC);
  cudaDeviceSynchronize();


  long nThreads = thread_num;
  for(int width = 2; width < index*2; width*=2){
    int arrayPerThread = index/(nThreads*width) + 1;
    mergeSort<<<1, thread_num>>>(output_datarange,temporary_array, index, width, arrayPerThread);
    temp_swap = output_datarange;
    output_datarange = temporary_array;
    temporary_array = temp_swap;
  }
  cudaDeviceSynchronize();
  // printf("ITEM : %d \n",output_datarange[0]);
  // printf("ITEM : %d \n",dev_c[0]);

  // Copy output vector from GPU buffer to host memory.
  cudaMemcpy(output, output_datarange, index * sizeof(int), cudaMemcpyDeviceToHost);

  cudaFree(output_datarange);
  cudaFree(temporary_array);
  cudaFree(dev_c);
  cudaFree(dev_a);
  cudaFree(dev_b);  
}

int get_median(int *n, int length)
{
  clock_t begin = clock();
  int mid = length / 2;
  if (length & 1)
    return n[mid];

  return (n[mid - 1] + n[mid]) / 2;
}

long get_floored_mean(int *n, int length)
{
  long sum = 0;
  for (int i = 0; i < length; i++)
  {
    sum += n[i];
  }

  return sum / length;
}

void print_array(int *n, int size)
{
  for (int i = 0; i < size; i++)
    printf("%d ", n[i]);
  printf("\n");
}


// call this function to start a nanosecond-resolution timer
struct timespec timer_start(){
  struct timespec start_time;
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start_time);
  return start_time;
}

// call this function to end a timer, returning nanoseconds elapsed as a long
double timer_end(struct timespec start_time){
  struct timespec end_time;
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end_time);
  double diffInNanos = (end_time.tv_sec - start_time.tv_sec) * (double)1e9 + (end_time.tv_nsec - start_time.tv_nsec);
  return diffInNanos;
}

int main(int argc, char **argv){
  struct timespec vartime = timer_start();
  int num_targets;
  int rowA,colA;
  int rowB,colB;

  char *ptr;

  int block_num = strtol(argv[1], &ptr, 10);
  if(block_num <= 0) block_num = 1;
  int thread_num = strtol(argv[2], &ptr, 10);
  if(thread_num <= 0) thread_num = 16;

  scanf("%d %d", &rowA, &colA);
  // INISIASI ARRAY/MATRIX
  int a[rowA*colA] = {0};
  // INPUT MATRIX KERNEL
  for (int i = 0; i < rowA; i++)
  {
    for (int j = 0; j < colA; j++)
    {
      scanf("%d", &a[i * colA + j]);
    }
  }

  scanf("%d %d %d", &num_targets, &rowB, &colB);
  // INISIASI ARRAY/MATRIX

  // printf("%d %d %d\n", num_targets, rowB, colB);
  int *b = (int *)malloc(num_targets*rowB*colB * sizeof(int));
  // INPUT MATRIX TARGET
  for (int k = 0; k < num_targets ; k++){
    for (int i = 0; i < rowB; i++)
    {
      for (int j = 0; j < colB; j++)
      {
        scanf("%d", &(b[(k*rowB*colB) + i * colB + j]));
      }
    }
  }
  
  int c[num_targets];

  convolutionWithCUDA(c, a, b, rowA,colA,rowB,colB,num_targets,block_num,thread_num);
  free(b);
  // print_array(c,num_targets);
  printf("\n");
  printf("MIN : %d \n",c[0]);
  printf("MAX : %d \n",c[num_targets-1]);
  printf("MEDIAN : %d \n",get_median(c,num_targets));
  printf("AVERAGE : %ld \n",get_floored_mean(c,num_targets));

  double time_elapsed_nanos = timer_end(vartime)/1000000000;
  printf("Time taken (second): %f \n", time_elapsed_nanos);

  cudaDeviceReset();
}